@@ -2,8 +2,7 @@
 # For details: https://github.com/PyCQA/astroid/blob/main/LICENSE
 # Copyright (c) https://github.com/PyCQA/astroid/blob/main/CONTRIBUTORS.txt
 
-"""this module contains a set of functions to handle inference on astroid trees
-"""
+"""This module contains a set of functions to handle inference on astroid trees."""
 
 from __future__ import annotations
 
@@ -11,10 +10,11 @@ import ast
 import functools
 import itertools
 import operator
+import typing
 from collections.abc import Callable, Generator, Iterable, Iterator
-from typing import TYPE_CHECKING, Any, TypeVar
+from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union
 
-from astroid import bases, decorators, helpers, nodes, protocols, util
+from astroid import bases, constraint, decorators, helpers, nodes, protocols, util
 from astroid.context import (
     CallContext,
     InferenceContext,
@@ -26,6 +26,7 @@ from astroid.exceptions import (
     AstroidError,
     AstroidIndexError,
     AstroidTypeError,
+    AstroidValueError,
     AttributeInferenceError,
     InferenceError,
     NameInferenceError,
@@ -45,21 +46,30 @@ if TYPE_CHECKING:
 # Prevents circular imports
 objects = util.lazy_import("objects")
 
-
-_FunctionDefT = TypeVar("_FunctionDefT", bound=nodes.FunctionDef)
-
-
-# .infer method ###############################################################
-
-
 _T = TypeVar("_T")
 _BaseContainerT = TypeVar("_BaseContainerT", bound=nodes.BaseContainer)
+_FunctionDefT = TypeVar("_FunctionDefT", bound=nodes.FunctionDef)
+
+GetFlowFactory = typing.Callable[
+    [
+        InferenceResult,
+        Optional[InferenceResult],
+        Union[nodes.AugAssign, nodes.BinOp],
+        InferenceResult,
+        Optional[InferenceResult],
+        InferenceContext,
+        InferenceContext,
+    ],
+    "list[functools.partial[Generator[InferenceResult, None, None]]]",
+]
+
+# .infer method ###############################################################
 
 
 def infer_end(
     self: _T, context: InferenceContext | None = None, **kwargs: Any
 ) -> Iterator[_T]:
-    """Inference's end for nodes that yield themselves on inference
+    """Inference's end for nodes that yield themselves on inference.
 
     These are objects for which inference does not have any semantic,
     such as Module or Consts.
@@ -69,15 +79,17 @@ def infer_end(
 
 # We add ignores to all assignments to methods
 # See https://github.com/python/mypy/issues/2427
-nodes.Module._infer = infer_end  # type: ignore[assignment]
-nodes.ClassDef._infer = infer_end  # type: ignore[assignment]
+nodes.Module._infer = infer_end
+nodes.ClassDef._infer = infer_end
 nodes.Lambda._infer = infer_end  # type: ignore[assignment]
 nodes.Const._infer = infer_end  # type: ignore[assignment]
 nodes.Slice._infer = infer_end  # type: ignore[assignment]
 
 
-def _infer_sequence_helper(node, context=None):
-    """Infer all values based on _BaseContainer.elts"""
+def _infer_sequence_helper(
+    node: _BaseContainerT, context: InferenceContext | None = None
+) -> list[SuccessfulInferenceResult]:
+    """Infer all values based on _BaseContainer.elts."""
     values = []
 
     for elt in node.elts:
@@ -140,7 +152,7 @@ def _update_with_replacement(
     lhs_dict: dict[SuccessfulInferenceResult, SuccessfulInferenceResult],
     rhs_dict: dict[SuccessfulInferenceResult, SuccessfulInferenceResult],
 ) -> dict[SuccessfulInferenceResult, SuccessfulInferenceResult]:
-    """Delete nodes that equate to duplicate keys
+    """Delete nodes that equate to duplicate keys.
 
     Since an astroid node doesn't 'equal' another node with the same value,
     this function uses the as_string method to make sure duplicate keys
@@ -165,7 +177,7 @@ def _update_with_replacement(
 def _infer_map(
     node: nodes.Dict, context: InferenceContext | None
 ) -> dict[SuccessfulInferenceResult, SuccessfulInferenceResult]:
-    """Infer all values based on Dict.items"""
+    """Infer all values based on Dict.items."""
     values: dict[SuccessfulInferenceResult, SuccessfulInferenceResult] = {}
     for name, value in node.items:
         if isinstance(name, nodes.DictUnpack):
@@ -182,7 +194,7 @@ def _infer_map(
             if any(not elem for elem in (key, safe_value)):
                 raise InferenceError(node=node, context=context)
             # safe_value is SuccessfulInferenceResult as bool(Uninferable) == False
-            values = _update_with_replacement(values, {key: safe_value})  # type: ignore[dict-item]
+            values = _update_with_replacement(values, {key: safe_value})
     return values
 
 
@@ -205,7 +217,7 @@ def _higher_function_scope(node: nodes.NodeNG) -> nodes.FunctionDef | None:
     while current.parent and not isinstance(current.parent, nodes.FunctionDef):
         current = current.parent
     if current and current.parent:
-        return current.parent  # type: ignore[return-value]
+        return current.parent  # type: ignore[no-any-return]
     return None
 
 
@@ -214,7 +226,7 @@ def infer_name(
     context: InferenceContext | None = None,
     **kwargs: Any,
 ) -> Generator[InferenceResult, None, None]:
-    """infer a Name: use name lookup rules"""
+    """Infer a Name: use name lookup rules."""
     frame, stmts = self.lookup(self.name)
     if not stmts:
         # Try to see if the name is enclosed in a nested function
@@ -229,13 +241,15 @@ def infer_name(
             )
     context = copy_context(context)
     context.lookupname = self.name
+    context.constraints[self.name] = constraint.get_constraints(self, frame)
+
     return bases._infer_stmts(stmts, context, frame)
 
 
 # pylint: disable=no-value-for-parameter
 # The order of the decorators here is important
 # See https://github.com/PyCQA/astroid/commit/0a8a75db30da060a24922e05048bc270230f5
-nodes.Name._infer = decorators.raise_if_nothing_inferred(  # type: ignore[assignment]
+nodes.Name._infer = decorators.raise_if_nothing_inferred(
     decorators.path_wrapper(infer_name)
 )
 nodes.AssignName.infer_lhs = infer_name  # won't work with a path wrapper
@@ -246,7 +260,7 @@ nodes.AssignName.infer_lhs = infer_name  # won't work with a path wrapper
 def infer_call(
     self: nodes.Call, context: InferenceContext | None = None, **kwargs: Any
 ) -> Generator[InferenceResult, None, InferenceErrorInfo]:
-    """infer a Call node by trying to guess what the function returns"""
+    """Infer a Call node by trying to guess what the function returns."""
     callcontext = copy_context(context)
     callcontext.boundnode = None
     if context is not None:
@@ -259,7 +273,10 @@ def infer_call(
         try:
             if hasattr(callee, "infer_call_result"):
                 callcontext.callcontext = CallContext(
-                    args=self.args, keywords=self.keywords, callee=callee
+                    args=self.args,
+                    keywords=self.keywords,
+                    callee=callee,
+                    parent_call_context=callcontext.callcontext,
                 )
                 yield from callee.infer_call_result(caller=self, context=callcontext)
         except InferenceError:
@@ -278,7 +295,7 @@ def infer_import(
     asname: bool = True,
     **kwargs: Any,
 ) -> Generator[nodes.Module, None, None]:
-    """infer an Import node: return the imported module/object"""
+    """Infer an Import node: return the imported module/object."""
     context = context or InferenceContext()
     name = context.lookupname
     if name is None:
@@ -293,7 +310,7 @@ def infer_import(
         raise InferenceError(node=self, context=context) from exc
 
 
-nodes.Import._infer = infer_import  # type: ignore[assignment]
+nodes.Import._infer = infer_import
 
 
 @decorators.raise_if_nothing_inferred
@@ -304,7 +321,7 @@ def infer_import_from(
     asname: bool = True,
     **kwargs: Any,
 ) -> Generator[InferenceResult, None, None]:
-    """infer a ImportFrom node: return the imported module/object"""
+    """Infer a ImportFrom node: return the imported module/object."""
     context = context or InferenceContext()
     name = context.lookupname
     if name is None:
@@ -339,7 +356,7 @@ def infer_attribute(
     context: InferenceContext | None = None,
     **kwargs: Any,
 ) -> Generator[InferenceResult, None, InferenceErrorInfo]:
-    """infer an Attribute node by using getattr on the associated object"""
+    """Infer an Attribute node by using getattr on the associated object."""
     for owner in self.expr.infer(context):
         if owner is util.Uninferable:
             yield owner
@@ -349,6 +366,11 @@ def infer_attribute(
         old_boundnode = context.boundnode
         try:
             context.boundnode = owner
+            if isinstance(owner, (nodes.ClassDef, bases.Instance)):
+                frame = owner if isinstance(owner, nodes.ClassDef) else owner._proxied
+                context.constraints[self.attrname] = constraint.get_constraints(
+                    self, frame=frame
+                )
             yield from owner.igetattr(self.attrname, context)
         except (
             AttributeInferenceError,
@@ -363,7 +385,7 @@ def infer_attribute(
 
 # The order of the decorators here is important
 # See https://github.com/PyCQA/astroid/commit/0a8a75db30da060a24922e05048bc270230f5
-nodes.Attribute._infer = decorators.raise_if_nothing_inferred(  # type: ignore[assignment]
+nodes.Attribute._infer = decorators.raise_if_nothing_inferred(
     decorators.path_wrapper(infer_attribute)
 )
 # won't work with a path wrapper
@@ -394,7 +416,7 @@ _SUBSCRIPT_SENTINEL = object()
 def infer_subscript(
     self: nodes.Subscript, context: InferenceContext | None = None, **kwargs: Any
 ) -> Generator[InferenceResult, None, InferenceErrorInfo | None]:
-    """Inference for subscripts
+    """Inference for subscripts.
 
     We're understanding if the index is a Const
     or a slice, passing the result of inference
@@ -431,6 +453,7 @@ def infer_subscript(
             except (
                 AstroidTypeError,
                 AstroidIndexError,
+                AstroidValueError,
                 AttributeInferenceError,
                 AttributeError,
             ) as exc:
@@ -512,7 +535,7 @@ def _infer_boolop(
     return InferenceErrorInfo(node=self, context=context)
 
 
-nodes.BoolOp._infer = _infer_boolop  # type: ignore[assignment]
+nodes.BoolOp._infer = _infer_boolop
 
 
 # UnaryOp, BinOp and AugAssign inferences
@@ -534,7 +557,7 @@ def _filter_operation_errors(
             # which shows that we can't infer the result.
             yield util.Uninferable
         else:
-            yield result  # type: ignore[misc]
+            yield result
 
 
 def _infer_unaryop(
@@ -608,10 +631,10 @@ def infer_unaryop(
 
 
 nodes.UnaryOp._infer_unaryop = _infer_unaryop
-nodes.UnaryOp._infer = infer_unaryop  # type: ignore[assignment]
+nodes.UnaryOp._infer = infer_unaryop
 
 
-def _is_not_implemented(const):
+def _is_not_implemented(const) -> bool:
     """Check if the given const node is NotImplemented."""
     return isinstance(const, nodes.Const) and const.value is NotImplemented
 
@@ -624,13 +647,14 @@ def _infer_old_style_string_formatting(
     TODO: Instead of returning Uninferable we should rely
     on the call to '%' to see if the result is actually uninferable.
     """
-    values = None
     if isinstance(other, nodes.Tuple):
         if util.Uninferable in other.elts:
             return (util.Uninferable,)
         inferred_positional = [helpers.safe_infer(i, context) for i in other.elts]
         if all(isinstance(i, nodes.Const) for i in inferred_positional):
             values = tuple(i.value for i in inferred_positional)
+        else:
+            values = None
     elif isinstance(other, nodes.Dict):
         values: dict[Any, Any] = {}
         for pair in other.items:
@@ -652,7 +676,14 @@ def _infer_old_style_string_formatting(
         return (util.Uninferable,)
 
 
-def _invoke_binop_inference(instance, opnode, op, other, context, method_name):
+def _invoke_binop_inference(
+    instance: InferenceResult,
+    opnode: nodes.AugAssign | nodes.BinOp,
+    op: str,
+    other: InferenceResult,
+    context: InferenceContext,
+    method_name: str,
+) -> Generator[InferenceResult, None, None]:
     """Invoke binary operation inference on the given instance."""
     methods = dunder_lookup.lookup(instance, method_name)
     context = bind_context_to_node(context, instance)
@@ -672,10 +703,21 @@ def _invoke_binop_inference(instance, opnode, op, other, context, method_name):
         raise InferenceError(node=method, context=context) from e
     if inferred is util.Uninferable:
         raise InferenceError
+    if not isinstance(
+        instance, (nodes.Const, nodes.Tuple, nodes.List, nodes.ClassDef, bases.Instance)
+    ):
+        raise InferenceError  # pragma: no cover # Used as a failsafe
     return instance.infer_binary_op(opnode, op, other, context, inferred)
 
 
-def _aug_op(instance, opnode, op, other, context, reverse=False):
+def _aug_op(
+    instance: InferenceResult,
+    opnode: nodes.AugAssign,
+    op: str,
+    other: InferenceResult,
+    context: InferenceContext,
+    reverse: bool = False,
+) -> functools.partial[Generator[InferenceResult, None, None]]:
     """Get an inference callable for an augmented binary operation."""
     method_name = protocols.AUGMENTED_OP_METHOD[op]
     return functools.partial(
@@ -689,7 +731,14 @@ def _aug_op(instance, opnode, op, other, context, reverse=False):
     )
 
 
-def _bin_op(instance, opnode, op, other, context, reverse=False):
+def _bin_op(
+    instance: InferenceResult,
+    opnode: nodes.AugAssign | nodes.BinOp,
+    op: str,
+    other: InferenceResult,
+    context: InferenceContext,
+    reverse: bool = False,
+) -> functools.partial[Generator[InferenceResult, None, None]]:
     """Get an inference callable for a normal binary operation.
 
     If *reverse* is True, then the reflected method will be used instead.
@@ -725,14 +774,20 @@ def _get_binop_contexts(context, left, right):
         yield new_context
 
 
-def _same_type(type1, type2):
+def _same_type(type1, type2) -> bool:
     """Check if type1 is the same as type2."""
     return type1.qname() == type2.qname()
 
 
 def _get_binop_flow(
-    left, left_type, binary_opnode, right, right_type, context, reverse_context
-):
+    left: InferenceResult,
+    left_type: InferenceResult | None,
+    binary_opnode: nodes.AugAssign | nodes.BinOp,
+    right: InferenceResult,
+    right_type: InferenceResult | None,
+    context: InferenceContext,
+    reverse_context: InferenceContext,
+) -> list[functools.partial[Generator[InferenceResult, None, None]]]:
     """Get the flow for binary operations.
 
     The rules are a bit messy:
@@ -766,8 +821,14 @@ def _get_binop_flow(
 
 
 def _get_aug_flow(
-    left, left_type, aug_opnode, right, right_type, context, reverse_context
-):
+    left: InferenceResult,
+    left_type: InferenceResult | None,
+    aug_opnode: nodes.AugAssign,
+    right: InferenceResult,
+    right_type: InferenceResult | None,
+    context: InferenceContext,
+    reverse_context: InferenceContext,
+) -> list[functools.partial[Generator[InferenceResult, None, None]]]:
     """Get the flow for augmented binary operations.
 
     The rules are a bit messy:
@@ -810,8 +871,14 @@ def _get_aug_flow(
     return methods
 
 
-def _infer_binary_operation(left, right, binary_opnode, context, flow_factory):
-    """Infer a binary operation between a left operand and a right operand
+def _infer_binary_operation(
+    left: InferenceResult,
+    right: InferenceResult,
+    binary_opnode: nodes.AugAssign | nodes.BinOp,
+    context: InferenceContext,
+    flow_factory: GetFlowFactory,
+) -> Generator[InferenceResult | util.BadBinaryOperationMessage, None, None]:
+    """Infer a binary operation between a left operand and a right operand.
 
     This is used by both normal binary operations and augmented binary
     operations, the only difference is the flow factory used.
@@ -892,7 +959,7 @@ def infer_binop(
 
 
 nodes.BinOp._infer_binop = _infer_binop
-nodes.BinOp._infer = infer_binop  # type: ignore[assignment]
+nodes.BinOp._infer = infer_binop
 
 COMPARE_OPS: dict[str, Callable[[Any, Any], bool]] = {
     "==": operator.eq,
@@ -1030,7 +1097,7 @@ def infer_augassign(
 
 
 nodes.AugAssign._infer_augassign = _infer_augassign
-nodes.AugAssign._infer = infer_augassign  # type: ignore[assignment]
+nodes.AugAssign._infer = infer_augassign
 
 # End of binary operation inference.
 
@@ -1054,8 +1121,8 @@ def infer_assign(
     context: InferenceContext | None = None,
     **kwargs: Any,
 ) -> Generator[InferenceResult, None, None]:
-    """infer a AssignName/AssignAttr: need to inspect the RHS part of the
-    assign node
+    """Infer a AssignName/AssignAttr: need to inspect the RHS part of the
+    assign node.
     """
     if isinstance(self.parent, nodes.AugAssign):
         return self.parent.infer(context)
@@ -1064,8 +1131,8 @@ def infer_assign(
     return bases._infer_stmts(stmts, context)
 
 
-nodes.AssignName._infer = infer_assign  # type: ignore[assignment]
-nodes.AssignAttr._infer = infer_assign  # type: ignore[assignment]
+nodes.AssignName._infer = infer_assign
+nodes.AssignAttr._infer = infer_assign
 
 
 @decorators.raise_if_nothing_inferred
@@ -1087,10 +1154,10 @@ def infer_empty_node(
 nodes.EmptyNode._infer = infer_empty_node  # type: ignore[assignment]
 
 
-def _populate_context_lookup(call, context):
+def _populate_context_lookup(call: nodes.Call, context: InferenceContext | None):
     # Allows context to be saved for later
     # for inference inside a function
-    context_lookup = {}
+    context_lookup: dict[InferenceResult, InferenceContext] = {}
     if context is None:
         return context_lookup
     for arg in call.args:
@@ -1108,7 +1175,7 @@ def _populate_context_lookup(call, context):
 def infer_ifexp(
     self: nodes.IfExp, context: InferenceContext | None = None, **kwargs: Any
 ) -> Generator[InferenceResult, None, None]:
-    """Support IfExp inference
+    """Support IfExp inference.
 
     If we can't infer the truthiness of the condition, we default
     to inferring both branches. Otherwise, we infer either branch
@@ -1176,4 +1243,4 @@ def infer_functiondef(
     return InferenceErrorInfo(node=self, context=context)
 
 
-nodes.FunctionDef._infer = infer_functiondef  # type: ignore[assignment]
+nodes.FunctionDef._infer = infer_functiondef
